#include <iostream>
#include <iomanip>
#include <string>
#include <time.h>
#include <fitsio.h>
#include "PSRFits.hh"
#include "Utils.hh"
#include "Block.hh"
#include "Header.hh"

using namespace Block;
using namespace Header;

namespace FITS
{
  
  PSRFits::PSRFits()
  {
    this->newFile = true;
    this->filestate = CLOSED;
  }
 //--------------------------------------------------------------------------

  PSRFits::PSRFits(std::string name)
  {
    this->filestate = CLOSED;
    this->openFile(name);
  }
 //--------------------------------------------------------------------------

  void PSRFits::setFITSname(Header::SURVEY surveyType)
  {
    /**
     * Generates a string containing the name of the FITS file, based on the
     *  values of the initial date and time of the first observations in the 
     *  file, and (if the multibeam receiver was used) the beam number. 
     * The format is determined by the survey type. Possible formats are:
     * <ul> <li> S70_yymmdd_hhmmss.fits
     *      <li> PM_yymmdd_hhmmss_bb.fits </ul>
     * \param surveyType The enumerated type of the survey data.
     * \return String containing the FITS filename.
     */
    std::string date;
    int j=0;
    for(int i=0;i<this->DATE_initial.size();i++)
      if(DATE_initial[i]!='/') date += this->DATE_initial[i];

    std::string time = removeLeadingBlanks(this->UT_initial);
    time = decToDMS( dmsToDec(time), 0, "");
    
    std::stringstream output;
    if(surveyType == S70) output << "S70_";
    else if(surveyType == PM) output << "PM_";
    else if(surveyType == PH) output << "PH_";
    else std::cerr << "Error <PSRFits::setFITSname> : unknown survey!\n";

    output << date << "_" << time;

    if((surveyType == PM)||(surveyType == PH))
      output << "_" 
	     << std::setw(2) << std::setfill('0') 
	     << this->beam_number;

    output << ".fits";

    this->FITSname = output.str();
    
  }
 //--------------------------------------------------------------------------

  bool PSRFits::fileExists()
  {
    /**
     * Checks to see whether the file is exists. Uses the CFITSIO function
     * fits_file_exists, and the parameter PSRFits::FITSname.
     * \return True if the file exists, false otherwise.
     */
    int status = 0, exists;
    fits_file_exists(this->FITSname.c_str(), &exists, &status);
    if(status) FITSError("PSRFits::fileExists",status);
    if(exists>0) return true;
    else return false;
  }
 //--------------------------------------------------------------------------

  int PSRFits::openFile(std::string name)
  {
    /**
     *   If the file exists, it is opened.
     *   If not, it is created using the correct template file.
     *    In the latter case, the fix function is called, and the file is
     *    flagged as "new".
     *  \param name Filename of the fits file to be opened.
     */
    int status = 0, exists;
    fits_file_exists(name.c_str(),&exists,&status);
    if(status) FITSError("PSRFits::openFile 1",status);
    status=0;
    if(exists<=0){
      char *template_file = (char *)getPSRFITSdefn().c_str();
      fits_create_template(&this->fptr, name.c_str(), template_file, &status);
      std::cerr << "Creating new file " << name << "\n";
    }
    else{
      this->newFile = false;
      fits_open_file(&this->fptr, name.c_str(), READWRITE, &status);
    }
    if(status) FITSError("PSRFits::openFile 2",status);
    else{
      if(exists<=0){
	this->newFile = true;
	this->fixPSRFITS();
      }
      this->gotoPHU();
      this->filestate = OPEN;
    }
    return status;
  }
 //--------------------------------------------------------------------------
  
  void PSRFits::addNewKeywords(DataBlock &block)
  {
    /**
     *   Adds new keywords to the PHU that have not yet been added into the 
     *   PSRFITS definition.
     *   Need to be in the PHU before calling this function.
     */

    int status = 0;
    
    // OBSFREQ
    double f0 = block.getKeyword<double>(FREQ_CHAN1);
    double bw = block.getKeyword<double>(CHANBW);
    double nc = block.getKeyword<double>(NCHAN);
    float fcent = f0 + ((nc+1)/2.) * (bw/nc);
    fits_update_key(this->fptr, TFLOAT, "OBSFREQ", &fcent, 
		    "[MHz] Observing frequency", &status);
    if(status) FITSError("PSRFits::addNewKeywords 1", status);

    // RA
    std::string ra  = deblank(block.getKeyword<std::string>(RA));
    status = 0;
    fits_update_key(this->fptr, TSTRING, "RA", (char *)ra.c_str(),
		    "Right Ascension", &status);
    if(status) FITSError("PSRFits::addNewKeywords 2", status);

    // DEC
    std::string dec = deblank(block.getKeyword<std::string>(DEC));
    status = 0;
    fits_update_key(this->fptr, TSTRING, "Dec", (char *)dec.c_str(),
		    "Declination", &status);
    if(status) FITSError("PSRFits::addNewKeywords 3", status);
    
    // DATE-OBS
    std::string dateobs = deblank(block.getKeyword<std::string>(DATE));
    if(dateobs[0] == '0') dateobs = "20" + dateobs;
    else                  dateobs = "19" + dateobs;
    for(int i=0;i<dateobs.size();i++) if(dateobs[i]=='/') dateobs[i]='-';
    dateobs += "T" + deblank(block.getKeyword<std::string>(UT));
    dateobs = dateobs.substr(0, dateobs.size()-7);
    status = 0;
    fits_update_key(this->fptr, TSTRING, "DATE-OBS", (char *)dateobs.c_str(),
		    "Date of observation (YY-MM-DDThh:mm:ss UTC)", &status);
    if(status) FITSError("PSRFits::addNewKeywords 4", status);

    // EQUINOX
    std::string equinox_str = block.getEquinox();
    if(equinox_str == Block::eq_b1950 || equinox_str == Block::eq_j2000){
      float equinox;
      if(equinox_str == Block::eq_b1950) equinox = 1950.;
      else equinox = 2000.;
      status = 0;
      fits_update_key(this->fptr, TFLOAT, "EQUINOX", &equinox,
		      "Equinox", &status);
      if(status) FITSError("PSRFits::addNewKeywords 5", status);
    }

    // BEAM INFORMATION
    float fval;
    // BMAJ
    fval = block.Survey().getBmaj();
    status = 0;
    fits_update_key(this->fptr, TFLOAT, "BMAJ", &fval, "", &status);
    if(status) FITSError("PSRFits::addNewKeywords 6", status);
    // BMIN
    fval = block.Survey().getBmin();
    status = 0;
    fits_update_key(this->fptr, TFLOAT, "BMIN", &fval, "", &status);
    if(status) FITSError("PSRFits::addNewKeywords 7", status);
    // BPA
    fval = block.Survey().getBpa();
    status = 0;
    fits_update_key(this->fptr, TFLOAT, "BPA", &fval, "", &status);
    if(status) FITSError("PSRFits::addNewKeywords 8", status);
    
  }
 //--------------------------------------------------------------------------
  
  int PSRFits::closeFile()
  {
    /**
     *   Closes the file if it is open. Adjusts the filestate variable as well.
     */
    int status = 0;
    if(this->filestate == OPEN){
      fits_close_file(this->fptr, &status);
      if(status) FITSError("PSRFits::closeFile",status);
      this->filestate = CLOSED;
    }
    return status;
  }
 //--------------------------------------------------------------------------
  
  int PSRFits::gotoPHU()
  {
    int status = 0;
    fits_movabs_hdu(this->fptr, 1, NULL, &status);    
    if(status) FITSError("PSRFits::gotoPHU",status);
    return status;
  }
 //--------------------------------------------------------------------------
  
  int PSRFits::gotoHistory()
  {
    int status = 0;
    char name[8];
    strcpy(name,"HISTORY");
    fits_movnam_hdu(this->fptr, BINARY_TBL, name, 0, &status);
    if(status) FITSError("PSRFits::gotoHistory",status);
    return status;
  }
 //--------------------------------------------------------------------------
  
  int PSRFits::gotoSubint()
  {
    int status = 0;
    char name[8];
    strcpy(name,"SUBINT");
    fits_movnam_hdu(this->fptr, BINARY_TBL, name, 0, &status);
    if(status) FITSError("PSRFits::gotoSubint",status);
    return status;
  }
 //--------------------------------------------------------------------------
  
  template <typename T> int PSRFits::updateKeyword(std::string name, T value)
  {
  /**
   *   Updates the keyword "name" in the current header unit, giving
   *    it the value given by "value".
   *   A templated function that should work for all cases.
   */
    int status=0;
    fits_update_key(this->fptr, TINT, name.c_str(), &value, NULL, &status);
    if(status) FITSError("PSRFits::updateKeyword",status);
    return status;
  }
  template <> int PSRFits::updateKeyword(std::string name, int value){
    int status=0;
    fits_update_key(this->fptr, TINT, (char *)name.c_str(), 
		    &value, NULL, &status);
    if(status) FITSError("PSRFits::updateKeyword",status);
    return status;
  }
  template <> int PSRFits::updateKeyword(std::string name, long value){
    int status=0;
    fits_update_key(this->fptr, TLONG, (char *)name.c_str(), 
		    &value, NULL, &status);
    if(status) FITSError("PSRFits::updateKeyword",status);
    return status;
  }
  template <> int PSRFits::updateKeyword(std::string name, float value){
    int status=0;
    fits_update_key(this->fptr, TFLOAT, (char *)name.c_str(),
		    &value, NULL, &status);
    if(status) FITSError("PSRFits::updateKeyword",status);
    return status;
  }
  template <> int PSRFits::updateKeyword(std::string name, double value){
    int status=0;
    fits_update_key(this->fptr, TDOUBLE, (char *)name.c_str(), 
		    &value, NULL, &status);
    if(status) FITSError("PSRFits::updateKeyword",status);
    return status;
  }
  template <> int PSRFits::updateKeyword(std::string name, std::string value){
    int status=0;
    fits_update_key(this->fptr, TSTRING, (char *)name.c_str(), 
		    (char *)value.c_str(), NULL, &status);
    if(status) FITSError("PSRFits::updateKeyword",status);
    return status;
  }
  template <> int PSRFits::updateKeyword(std::string name, char *value){
    int status=0;
    fits_update_key(this->fptr, TSTRING, (char *)name.c_str(), 
		    value, NULL, &status);
    if(status) FITSError("PSRFits::updateKeyword",status);
    return status;
  }
 //--------------------------------------------------------------------------

  int PSRFits::getColnum(char *name)
  {
    int status=0,colnum=0;
    fits_get_colnum(this->fptr, CASEINSEN, name, &colnum, &status);
    if(status) FITSError("getColnum", status);
    return colnum;
  }
  
  long PSRFits::getNumRows()
  {
    int status=0;
    long nrows=0;
    fits_get_num_rows(this->fptr, &nrows, &status);
    if(status) FITSError("getNumRows", status);
    return nrows;
  }
 //--------------------------------------------------------------------------
  
  template <class T> int PSRFits::writeColumnValue(char *name, T value)
  {
    /**
     *   Writes a new value in the column "name" in the current binary table.
     *   It adds a single value at the bottom of the table (ie. in the largest
     *    numbered row).
     *   A templated function that should work for all cases.
     */
    int status=0;
    int colnum = this->getColnum(name);
    long nrows = this->getNumRows();
    fits_write_col(this->fptr, TINT, colnum, nrows, 1, 1, &value, &status);
    if(status) FITSError("PSRFits::writeColumnValue", status);
    return status;
  }
  template <> int PSRFits::writeColumnValue(char *name, long value)
  {
    int status=0;
    int colnum = this->getColnum(name);
    long nrows = this->getNumRows();
    fits_write_col(this->fptr, TLONG, colnum, nrows, 1, 1, &value, &status);
    if(status) FITSError("PSRFits::writeColumnValue", status);
    return status;
  }
  template <> int PSRFits::writeColumnValue(char *name, float value)
  {
    int status=0;
    int colnum = this->getColnum(name);
    long nrows = this->getNumRows();
    fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 1, &value, &status);
    if(status) FITSError("PSRFits::writeColumnValue", status);
    return status;
  }
  template <> int PSRFits::writeColumnValue(char *name, double value)
  {
    int status=0;
    int colnum = this->getColnum(name);
    long nrows = this->getNumRows();
    fits_write_col(this->fptr, TDOUBLE, colnum, nrows, 1, 1, &value, &status);
    if(status) FITSError("PSRFits::writeColumnValue", status);
    return status;
  }
  template <> int PSRFits::writeColumnValue(char *name, char *value)
  {
    int status=0;
    int colnum = this->getColnum(name);
    long nrows = this->getNumRows();
    fits_write_col(this->fptr, TSTRING, colnum, nrows, 1, 1, &value, &status);
    if(status) FITSError("PSRFits::writeColumnValue", status);
    return status;
  }
 //--------------------------------------------------------------------------

  int PSRFits::startNewRow()
  {
    /**
     *    Adds a new row to the bottom of the current binary table.
     *    Also adds the row number as the ISUBINT column value.
     */
    int nrows = this->getNumRows();
    int status = 0;
    fits_insert_rows(this->fptr, nrows, 1, &status);
    if(status) FITSError("PSRFits::startNewRow", status);
    return status;
  }
 //--------------------------------------------------------------------------

  void PSRFits::startNewSubintRow()
  {
    /**
     *    Adds a new row to the bottom of the current binary table.
     *    Also adds the row number as the ISUBINT column value.
     */
    this->startNewRow();
    this->writeColumnValue("ISUBINT", this->getNumRows());
   }
 //--------------------------------------------------------------------------


}
