#ifndef PSRFITS_HH_
#define PSRFITS_HH_
#include <fitsio.h>
#include <string>
#include "Block.hh"
#include "Header.hh"

using namespace Block;
using namespace Header;

/**
 * The namespace to control all things related to reading and writing
 * fits files in the PSRFITS format.
 */ 

namespace FITS
{

  /** Has a function worked or not? */
  enum RESULT { SUCCESS, FAILURE };
	
  /** The state of the file */
  enum STATE { OPEN,  ///< file is open
	       CLOSED ///< file is closed
  };
  
  /** 
   * Report a CFITSIO error, along with the offending function.  
   * An error message is printed, along with the relevant CFITSIO error
   * code and message.  
   * \param function The function from which the error was called.  
   * \param status The status value as returned by the CFITSIO function. 
   *  Generally, if status != 0, an error has occurred.
   */
  inline void FITSError(std::string function, int status){
    std::cerr << "\aFITSError <" << function << "> (status="<<status<<"): ";
    fits_report_error(stderr, status);
  }

  //===========================================================

  /**
   *  A class to keep track of the header keywords for the SUBINT
   *  binary table extension.
   */
  class Subint
  {
  public:


    Subint(){
      allocated = false;
      nch_strt = 0;
      int_type = "TIME";
      int_unit = "SEC";
    };
    virtual ~Subint(){};

    friend class PSRFits;

    long   getNch_file(){return nch_file;};
    void   setNch_file(long l){nch_file=l;};
    long   getNch_strt(){return nch_strt;};
    void   setNch_strt(long l){nch_strt=l;};
    long   getNbin(){return nbin;};
    void   setNbin(long l){nbin=l;};
    long   getNpol(){return npol;};
    void   setNpol(long l){npol=l;};
    long   getNsblk(){return nsblk;};
    void   setNsblk(long l){nsblk=l;};
    float  getTsamp(){return tsamp;};
    void   setTsamp(float f){tsamp=f;};
    std::string getInt_type(){return int_type;};
    void   setInt_type(std::string s){int_type=s;};
    std::string getInt_unit(){return int_unit;};
    void   setInt_unit(std::string s){int_unit=s;};
    bool   getAllocatedFlag(){return allocated;};
    void   setAllocatedFlag(bool b){allocated=b;};

    float *getDAT_freq(){return dat_freq;};
    void   setDAT_freq(float *f){dat_freq=f;};
    float *getDAT_wts(){return dat_wts;};
    void   setDAT_wts(float *f){dat_wts=f;};
    float *getDAT_offs(){return dat_offs;};
    void   setDAT_offs(float *f){dat_offs=f;};
    float *getDAT_scl(){return dat_scl;};
    void   setDAT_scl(float *f){dat_scl=f;};

    /** A function that will write to a file the full set of keywords. */
    void writeKeywords(fitsfile *fptr);
 
  private:
    long        nch_strt;   ///< The SUBINT table keyword NCH_STRT
    long        nch_file;   ///< The SUBINT table keyword NCH_FILE
    long        npol;	    ///< The SUBINT table keyword NPOL
    long        nsblk;	    ///< The SUBINT table keyword NSBLK
    std::string int_type;   ///< The SUBINT table keyword INT_TYPE
    std::string int_unit;   ///< The SUBINT table keyword INT_UNIT
    float       tsamp;	    ///< The SUBINT table keyword TSAMP
    long        nbin;       ///< The SUBINT table keyword NBIN
    bool        allocated;  ///< Have the keywords been allocated yet?

    float      *dat_freq;   ///< The SUBINT table array DAT_FREQ
    float      *dat_wts;    ///< The SUBINT table array DAT_WTS
    float      *dat_offs;   ///< The SUBINT table array DAT_OFFS
    float      *dat_scl;    ///< The SUBINT table array DAT_SCL


  };

  
  const int SUBINT_kw_nch_strt = 0; 
  ///< The nch_strt keyword for the SUBINT table.

  const std::string SUBINT_kw_int_type = "TIME";
  ///< The int_type keyword for the SUBINT table.

  const std::string SUBINT_kw_int_unit = "SEC"; 
  ///< The int_unit keyword for the SUBINT table.
  
  const std::string PHU_kw_coord_md = "J2000.0"; 
  ///< The coord_md keyword for the Primary header.
  
  /**
   * A class to deal with individual fits files.
   * This keeps track of all information relevant for a particular PSRFITS file.
   * All the functions that write to particular extensions are in here, as 
   *  well as general I/O routines.
   */ 

  class PSRFits
  {
  public:
    PSRFits(); ///<
    PSRFits(std::string name);
    virtual ~PSRFits(){};

    // Inline functions
    /** Is the file newly created? */
    bool        isNewFile(){return newFile;};

    /** Is the file open or closed? */
    STATE       getFilestate(){return filestate;};

    /** Set the file state. */
    void        setFilestate(STATE S){filestate=S;};

    /** Get the UT of the first observations in the file. */
    std::string getUT(){return UT_initial;};

    /** Save the UT of the first observations. */
    void        setUT(std::string s){UT_initial=s;};

    /** Get the date of the first observations in the file. */
    std::string getDate(){return DATE_initial;};

    /** Save the date of the first observations. */
    void        setDate(std::string s){DATE_initial=s;};

    /** Get the beam number used for the observations. */
    int         getBeam(){return beam_number;};

    /** Save the beam number used for the observations. */
    void        setBeam(int i){beam_number=i;};

    /** Get the filename of the fits file. */
    std::string getFITSname(){return FITSname;};

    /** Get the number of the first block stored in the file. */
    int         getInitialBlock(){return blk_initial;};

    /** Save the number of the first block stored in the file. */
    void        setInitialBlock(int block){blk_initial = block;};

    /** Opens file -- uses openFile(std::string) */
    int openFile(){ return this->openFile(this->FITSname); };


    // Functions found in PSRFits.cc (in the same order...)
    void setFITSname(Header::SURVEY surveyType); 
         ///< Generate the name of the fits file.
    bool fileExists();                           
         ///< Does the file exist on disk already?
    int  openFile(std::string name);              
         ///< Opens the file named by the argument.
    void fixPSRFITS();                           
         ///< Makes corrections to PSRFITS definition.
    void addNewKeywords(DataBlock &block);       
         ///< Writes new keywords to the primary fits header.
    int  closeFile();   ///< Closes the file.
    int  gotoPHU();     ///< Moves to the primary header unit
    int  gotoHistory(); ///< Moves to the HISTORY binary table extension.
    int  gotoSubint();  ///< Moves to the SUBINT binary table extension.

    template <class T> int updateKeyword(std::string name, T value); 
         ///< Update the value of a keyword.

    int  getColnum(char *name); 
         ///< Returns the number of the named column.
    long getNumRows();          
         ///< Returns the number of rows in the current binary table.
    template <class T> int writeColumnValue(char *name, T value);
         ///< Write a value to a named column.

    int  startNewRow();         
         ///< Add new row to current table.
    void startNewSubintRow();   
         ///< Add new row to SUBINT table.

    void writePHU(DataBlock &block);             
         ///< Write Primary Header Unit data.
    int  initialise(DataBlock &block);
    void setupSubint(DataBlock &block);
    int  writeSubint(DataBlock &block);          
    void writeSubintRow(DataBlock &block);          
         ///< Write SUBINT table data.
    void writeHistory(int argc, char * argv[]); 
         ///< Write information to HISTORY table.
    void writeSupportArrays(DataBlock &block); 
         ///< Write multi-valued columns in SUBINT table.
    void writeDataArray(DataBlock &block); 
         ///< Write data to SUBINT table.

  private:
    fitsfile *fptr;           ///< The CFITSIO file variable
    std::string FITSname;     ///< The output FITS filename
    bool newFile;             ///< Is the file a new one?
    STATE filestate;          ///< OPEN or CLOSED?
    std::string UT_initial;   ///< The UT at the start of the FITS file data.
    std::string DATE_initial; ///< The UT date at the start of the data.
    int beam_number;          ///< What beam are we using? Necessary for FITS file name
    int blk_initial;          ///< The block number of the first block in the file.
    Subint subintTable;       ///< Subint class containing table header keywords.
  };

}
#endif /*PSRFITS_HH_*/
