#include <iostream>
#include <iomanip>
#include <string>
#include <time.h>
#include <fitsio.h>
#include "PSRFits.hh"
#include "Utils.hh"
#include "Block.hh"
#include "Header.hh"

#define USE_X true

using namespace Block;
using namespace Header;

namespace FITS
{

  void PSRFits::fixPSRFITS()
  {
    /**
     *  Corrects known errors that exist in the PSRFITS definition:
     *  <ul><li> Set the TFORM keyword for the data column in the SUBINT table
     *      <li> Fix the units for the RA & DEC columns
     *  </ul>
     */
    if(this->gotoSubint()){
      std::cerr << 
	"Error <PSRFits::fixPSRFITS> : Could not go to the SUBINT table.\n";
      return;
    }
    
    int status=0;
    char cval[9];
    if(USE_X) strcpy(cval,"16X");
    else      strcpy(cval,"2B");
    fits_update_key(this->fptr, TSTRING, "TFORM19", cval, NULL, &status);
    if(status) FITSError("PSRFits::fixPSRFITS 1", status);

//     status=0;
//     strcpy(cval,"deg");
//     fits_update_key(this->fptr, TSTRING, "TUNIT6", cval, NULL, &status);
//     if(status) FITSError("PSRFits::fixPSRFITS 2", status);
//     status=0;
//     fits_update_key(this->fptr, TSTRING, "TUNIT7", cval, NULL, &status);
//     if(status) FITSError("PSRFits::fixPSRFITS 3", status);

//     status=0;
//     strcpy(cval,"MHz");
//     fits_update_key(this->fptr, TSTRING, "TUNIT15", cval, NULL, &status);
//     if(status) FITSError("PSRFits::fixPSRFITS 4", status);

    this->gotoPHU();
  }

 //--------------------------------------------------------------------------

  void PSRFits::writePHU(DataBlock &block)
  {
    /**
     *   Writes the header information contained in the DataBlock block
     *    to the primary header unit.
     *   At the end the current header unit will be the PHU.
     */
    this->gotoPHU();

    char *todaysdate     = new char[25];
    int timeref,status=0;
    fits_get_system_time(todaysdate, &timeref, &status);
    if(status) FITSError("PSRFits::writePHU", status);
    std::string projid   = block.getSurveyName();
    std::string telescop = block.getKeyword<std::string>(TELID);
    // CHECK -- the following PHU keywords are all survey dependent
    // CHECK -- need to find correct values.
    double ant_x         = block.Survey().getAntX();
    double ant_y         = block.Survey().getAntY();
    double ant_z         = block.Survey().getAntZ();
    std::string frontend = block.Survey().getFrontend();
    std::string fd_poln  = block.Survey().getFd_poln();
    int fd_hand          = block.Survey().getFd_hand();
    float fd_sang        = block.Survey().getFd_sang();
    float fd_xyph        = block.Survey().getFd_xyph();
    std::string backend  = block.Survey().getBackend();
    std::string beconfig = block.Survey().getBeconfig();
    int be_phase         = block.Survey().getBe_phase();
    int be_dcc           = block.Survey().getBe_dcc();
    double tcycle        = block.Survey().getTcycle();
    int nrcvr            = block.getKeyword<int>(NFILTER);
    std::string obs_mode = block.Survey().getObs_mode();
    std::string src_name = deblank(block.getKeyword<std::string>(PSRNAME));
//     std::string coord_md = PHU_kw_coord_md;
    std::string coord_md = block.getEquinox();
    //  CHECK -- NEED TO DECIDE ON RA/DEC OR GL/GB
    std::string stt_crd1 = deblank(block.getKeyword<std::string>(RA));
    std::string stt_crd2 = deblank(block.getKeyword<std::string>(DEC));
    std::string trk_mode = block.Survey().getTrk_mode();
    std::string stp_crd1 = deblank(block.getKeyword<std::string>(RA));
    std::string stp_crd2 = deblank(block.getKeyword<std::string>(DEC));
    float scanlen        = block.Survey().getScanlen();
    std::string fd_mode  = deblank(block.getKeyword<std::string>(FDCNTRL));
    float fa_req         = block.getKeyword<float>(FEED_ANGLE);
    std::string cal_mode = block.Survey().getCal_mode();
    float cal_freq       = block.Survey().getCal_freq();
    float cal_dcyc       = block.Survey().getCal_dcyc();
    float cal_phs        = block.Survey().getCal_phs();
    // CHECK -- STT_DATE NEEDS TO BE PUT INTO A STANDARD FORMAT.
    std::string stt_date = deblank(block.getKeyword<std::string>(DATE));
    if(stt_date[0] == '0') stt_date = "20" + stt_date;
    else                   stt_date = "19" + stt_date;
    for(int i=0;i<stt_date.size();i++) if(stt_date[i]=='/') stt_date[i]='-';
    std::string stt_time = deblank(block.getKeyword<std::string>(UT));
    long stt_imjd        = block.getKeyword<int>(MJD);
    long stt_smjd        = long(dmsToDec(stt_time) * 3600.);
    double stt_offs      = 0.; //CHECK -- what is STT_OFFS?
    double stt_lst       = dmsToDec(block.getKeyword<std::string>(LST))*3600.;

//     this->UT_initial = stt_time;

    this->updateKeyword("DATE",     todaysdate);
    this->updateKeyword("PROJID",   projid);
    this->updateKeyword("TELESCOP", telescop);
    this->updateKeyword("ANT_X",    ant_x);          
    this->updateKeyword("ANT_Y",    ant_y);        
    this->updateKeyword("ANT_Z",    ant_z);        
    this->updateKeyword("FRONTEND", frontend);
    this->updateKeyword("FD_POLN",  fd_poln);
    this->updateKeyword("FD_HAND",  fd_hand);
    this->updateKeyword("FD_SANG",  fd_sang);
    this->updateKeyword("FD_XYPH",  fd_xyph);
    this->updateKeyword("BACKEND",  backend);
    this->updateKeyword("BECONFIG", beconfig);
    this->updateKeyword("BE_PHASE", be_phase);
    this->updateKeyword("BE_DCC",   be_dcc);
    this->updateKeyword("TCYCLE",   tcycle);
    this->updateKeyword("NRCVR",    nrcvr);
    this->updateKeyword("OBS_MODE", obs_mode);
    this->updateKeyword("SRC_NAME", src_name);
    this->updateKeyword("COORD_MD", coord_md);
    this->updateKeyword("STT_CRD1", stt_crd1);
    this->updateKeyword("STT_CRD2", stt_crd2);
    this->updateKeyword("TRK_MODE", trk_mode);
    this->updateKeyword("STP_CRD1", stp_crd1);
    this->updateKeyword("STP_CRD2", stp_crd2);
    this->updateKeyword("SCANLEN",  scanlen);
    this->updateKeyword("FD_MODE",  fd_mode);
    this->updateKeyword("FA_REQ",   fa_req);
    this->updateKeyword("CAL_MODE", cal_mode);
    this->updateKeyword("CAL_FREQ", cal_freq);
    this->updateKeyword("CAL_DCYC", cal_dcyc);
    this->updateKeyword("CAL_PHS ", cal_phs);
    this->updateKeyword("STT_DATE", stt_date);
    this->updateKeyword("STT_TIME", stt_time);
    this->updateKeyword("STT_IMJD", stt_imjd);
    this->updateKeyword("STT_SMJD", stt_smjd);
    this->updateKeyword("STT_OFFS", stt_offs);
    this->updateKeyword("STT_LST ", stt_lst);

    this->addNewKeywords(block);

  }
 //--------------------------------------------------------------------------

  void PSRFits::writeHistory(int argc, char * argv[])
  {
    /**
     * Write processing history information to the HISTORY binary table 
     * extension. This is just a row of today's date and what was entered
     * in the command line.
     * \param argc The number of command line arguments.
     * \param argv Each command line argument.
     */
    
   this->gotoHistory();
   char *todaysdate     = new char[25];
   int timeref,status=0;
   fits_get_system_time(todaysdate, &timeref, &status);
   if(status) FITSError("PSRFits::writePHU", status);

   std::stringstream historyMessage;
   historyMessage << argv[0] << " " << argv[1];

   this->startNewRow(); 
   this->writeColumnValue("DATE_PRO", todaysdate);
   this->writeColumnValue("PROC_CMD", (char *)historyMessage.str().c_str());
   
   this->gotoPHU();

  }
 //--------------------------------------------------------------------------

  int PSRFits::writeSubint(DataBlock &block)
  {
    /**
     *   Writes the header information and data contained in the DataBlock 
     *    block to the SUBINT binary table extension.
     *   This does the following:
     *   <ul><li> Update the table headers
     *       <li> Start a new row.
     *       <li> Update the single-valued columns for the current row.
     *       <li> Call writeSupportArrays to update the array-valued columns.
     *       <li> Call writeDataArray to write the data column.
     *   </ul>
     *   At the end the current header unit will be the SUBINT extension.
     */
    if(this->gotoSubint()){
      std::cerr << 
	"Error <PSRFits::writeSubint> : Could not go to the SUBINT table.\n";
      return 1;
    }
    
    this->updateKeyword("NCH_FILE", block.getKeyword<int>(NCHAN));
    this->updateKeyword("INT_TYPE", SUBINT_kw_int_type);
    this->updateKeyword("INT_UNIT", SUBINT_kw_int_unit);
    this->updateKeyword("NCH_STRT", SUBINT_kw_nch_strt);

    this->startNewSubintRow(); 
    // CHECK -- need some check of blk_cntr to make sure the number is 
    // CHECK -- incrementing correctly?
    float tsubint = block.getKeyword<float>(BLK_SEC);
    double off_sub;
    // CHECK -- OFF_SUB is the offset, from the start (taken to mean the 
    // CHECK -- starting time listed in the PHU), of the centre of the 
    // CHECK -- current subintegration.

//     off_sub = dmsToDec(block.getKeyword<std::string>(UT)) - 
//       dmsToDec(this->UT_initial);
//     if(off_sub<0.) off_sub += 24.;
//     off_sub = off_sub*3600. + block.getKeyword<float>(BLK_SEC)/2.;
    
    off_sub = 0.5*(2.*(block.getKeyword<int>(BLK_CNTR)-this->blk_initial)+1)
      * block.getKeyword<double>(BLK_SEC);

    double lst_sub  = dmsToDec(block.getKeyword<std::string>(LST));
    double ra_sub   = dmsToDec(block.getKeyword<std::string>(RA))*15.;
    double dec_sub  = dmsToDec(block.getKeyword<std::string>(DEC));
    double glon_sub = block.getKeyword<double>(GL);
    double glat_sub = block.getKeyword<double>(GB);
    float  fd_ang   = block.getKeyword<float>(FEED_ANGLE);
    float  par_ang  = block.getKeyword<float>(PANGLE);
    float  pos_ang  = remainderf(par_ang + fd_ang, 360.);
    float  tel_az   = block.getKeyword<float>(TEL_AZ);
    float  tel_zen  = block.getKeyword<float>(TEL_ZEN);
    
    this->writeColumnValue("TSUBINT",  tsubint);
    this->writeColumnValue("OFFS_SUB", off_sub);
    this->writeColumnValue("LST_SUB",  lst_sub);
    this->writeColumnValue("RA_SUB",   ra_sub);
    this->writeColumnValue("DEC_SUB",  dec_sub);
    this->writeColumnValue("GLON_SUB", glon_sub);
    this->writeColumnValue("GLAT_SUB", glat_sub);
    this->writeColumnValue("FD_ANG",   fd_ang);
    this->writeColumnValue("POS_ANG",  pos_ang);
    this->writeColumnValue("PAR_ANG",  par_ang);
    this->writeColumnValue("TEL_AZ",   tel_az);
    this->writeColumnValue("TEL_ZEN",  tel_zen);
    
    this->writeSupportArrays(block); //DAT_FREQ,DAT_WTS,DAT_OFFS,DAT_SCL

    this->writeDataArray(block);
   
    this->gotoPHU();

    return 0;

  }
 //--------------------------------------------------------------------------

  void PSRFits::writeSupportArrays(DataBlock &block)
  {
    /**
     *   A function to write the multi-valued columns to the SUBINT table.
     *   These columns are:
     *   <ul><li> DAT_FREQ: the frequencies of each channel
     *       <li> DAT_WTS:  all set to 1. 
     *       <li> DAT_OFFS: all set to 0.
     *       <li> DAT_SCL:  all set to 1.
     *   </ul>
     *   The current header at the end is still SUBINT.
     */
    int numChannels = block.getKeyword<int>(NCHAN);
    float *dat_freq = new float[numChannels];
    dat_freq[0] = block.getKeyword<float>(FREQ_CHAN1);
    float deltaFreq = block.getKeyword<float>(CHANBW) / float(numChannels);
    for(int i=1;i<numChannels;i++) dat_freq[i] = dat_freq[i-1] + deltaFreq;

    float *dat_wts = new float[numChannels];
    for(int i=0;i<numChannels;i++) dat_wts[i] = 1.;
    float *dat_offs = new float[numChannels];
    for(int i=0;i<numChannels;i++) dat_offs[i] = 0.;
    float *dat_scl = new float[numChannels];
    for(int i=0;i<numChannels;i++) dat_scl[i] = 1.;

    int status = 0;
    long nrows = this->getNumRows();

    int colnum = this->getColnum("DAT_FREQ");
    status=0;
    fits_modify_vector_len (this->fptr, colnum, numChannels, &status);
    if(status) FITSError("PSRFits::writeSupportArray", status);
    status=0;
    fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		   numChannels, dat_freq, &status);
    if(status) FITSError("PSRFits::writeSupportArrays", status);

    colnum = this->getColnum("DAT_WTS");
    status=0;
    fits_modify_vector_len (this->fptr, colnum, numChannels, &status);
    if(status) FITSError("PSRFits::writeSupportArray", status);
    status=0;
    fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		   numChannels, dat_wts, &status);
    if(status) FITSError("PSRFits::writeSupportArrays", status);

    colnum = this->getColnum("DAT_OFFS");
    status=0;
    fits_modify_vector_len (this->fptr, colnum, numChannels, &status);
    if(status) FITSError("PSRFits::writeSupportArray", status);
    status=0;
    fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		   numChannels, dat_offs, &status);
    if(status) FITSError("PSRFits::writeSupportArrays", status);

    colnum = this->getColnum("DAT_SCL");
    status=0;
    fits_modify_vector_len (this->fptr, colnum, numChannels, &status);
    if(status) FITSError("PSRFits::writeSupportArray", status);
    status=0;
    fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		   numChannels, dat_scl, &status);
    if(status) FITSError("PSRFits::writeSupportArrays", status);

  }
 //--------------------------------------------------------------------------

  void PSRFits::writeDataArray(DataBlock &block)
  {
    /**
     *   A function to write the multi-dimensional data column to the SUBINT 
     *    binary table.
     *   
     *   The current header at the end is still SUBINT.
     */
    long nrows = this->getNumRows();
    int colnum = this->getColnum("DATA");
    int status = 0;

    int numBin = 1;

    int numChan = block.getKeyword<int>(NCHAN,0);

    int numPol;
    if(block.getKeyword<int>(NBITS,1)>0) numPol=2;
    else numPol = 1;

    int numSampBlock  = block.getKeyword<int>(SAMP_GRP,0) * 
      block.getKeyword<int>(GRP_BLK);

    int dataPerByte = 8 / block.getKeyword<int>(NBITS,0);


    int numData = numChan * numPol * numSampBlock;
    //CHECK -- IS THIS CONVERSION TO BYTES CORRECT?
    int numBytes = numData / dataPerByte;   

    unsigned char *data = (unsigned char *)block.getData();

    //Fixing size of vector
    if(this->newFile){
      status=0;
      if(USE_X) fits_modify_vector_len (this->fptr, colnum, numData, &status);
      else      fits_modify_vector_len (this->fptr, colnum, numBytes, &status);
      if(status) FITSError("PSRFits::writeDataArray 1", status);
    }
 
    long naxis=4,naxes[4];
    // CHECK -- ORDER OF AXES: first axis is quickest-varying one, so this is 
    // CHECK --  the channel axis. Next quickest is the pol axis (equivalent 
    // CHECK --  to the filter? CHECK THIS). Slowest varying is the sample  
    // CHECK --  axis (which has length numBin).
    // CHECK -- IS THE CONVERSION TO BYTES CORRECT? We are dividing the channel
    // CHECK --  axis, as this should be divisible by 8.
    naxes[0] = numBin;

    if(USE_X) naxes[1] = numChan; 
    else      naxes[1] = numChan / dataPerByte; 

    naxes[2] = numPol;

    naxes[3] = numSampBlock;

    if(this->newFile){
      status=0;
      // delete the existing "TDIM19 = (*,*,*,*)"
      fits_delete_key(this->fptr, "TDIM19", &status);
      if(status) FITSError("PSRFits::writeDataArray 2", status);
      status=0;
      fits_write_tdim(this->fptr, colnum, naxis, naxes, &status);
      if(status) FITSError("PSRFits::writeDataArray 3", status);
    }

    status=0;
    fits_write_col_byt(this->fptr, colnum, nrows, 1, numBytes, data, &status);
    if(status) FITSError("PSRFits::writeDataArray 4", status);

//     if(USE_X && this->newFile) {
    if(this->newFile && this->getNumRows()>1) {
      this->newFile = false;
      status=0;
      fits_delete_rows(this->fptr, nrows+1, 1, &status);
      if(status) FITSError("PSRFits::writeDataArray 5", status);
    }

  }
 //--------------------------------------------------------------------------

  int PSRFits::initialise(DataBlock &block)
  {
    /**
     * Initialises a fits file by setting the start variables,
     *  calling the PHU setup, and setting the header keywords for
     *  the subint table.
     */

    this->setUT(block.getKeyword<std::string>(UT_STRT));

    this->setDate(block.getKeyword<std::string>(DATE));

    this->setBeam(block.getKeyword<int>(IBEAM));

    this->setFITSname(block.getSurveyType());

    this->setInitialBlock(block.getKeyword<int>(BLK_CNTR));

    if(this->fileExists()){
      std::cerr << "\aThe file "<< this->getFITSname() 
		<< " already exists.\nDon't want to overwrite, so exiting...\n";
      return FAILURE;
    }
    if(this->openFile()) {
      std::cerr << "\aCannot open file "<< this->getFITSname()
		<< ".\nExiting...\n";
      return FAILURE;
    }

    this->writePHU(block);

    this->setupSubint(block);
    

    return SUCCESS;

  }
 //--------------------------------------------------------------------------

  void PSRFits::setupSubint(DataBlock &block)
  {
    
    this->subintTable.nbin = 1; //CHECK --> needs to be made more general!
    this->subintTable.nch_file = block.getKeyword<int>(NCHAN);
    this->subintTable.npol = 1;
    this->subintTable.nsblk = block.getKeyword<int>(SAMP_GRP,0) * 
      block.getKeyword<int>(GRP_BLK);
    this->subintTable.tsamp = block.getKeyword<float>(SAMP_INT,0)/1000.;

    // define the DAT_FREQ array;
    float deltaFreq = block.getKeyword<float>(CHANBW) / 
      float(this->subintTable.nch_file);
    this->subintTable.dat_freq = new float[this->subintTable.nch_file];
    this->subintTable.dat_freq[0] = block.getKeyword<float>(FREQ_CHAN1);
    for(int i=1;i<this->subintTable.nch_file;i++) 
      this->subintTable.dat_freq[i] = this->subintTable.dat_freq[i-1] + deltaFreq;

    // define the DAT_WTS array;
    this->subintTable.dat_wts = new float[this->subintTable.nch_file];
    for(int i=0;i<this->subintTable.nch_file;i++) this->subintTable.dat_wts[i] = 1.;

    // define the DAT_OFFS array;
    this->subintTable.dat_offs = new float[this->subintTable.nch_file];
    for(int i=0;i<this->subintTable.nch_file;i++) this->subintTable.dat_offs[i] = 0.;

    // define the DAT_SCL array;
    this->subintTable.dat_scl = new float[this->subintTable.nch_file];
    for(int i=0;i<this->subintTable.nch_file;i++) this->subintTable.dat_scl[i] = 1.;
    
    
    this->subintTable.setAllocatedFlag(true);

    // now write the keywords to the subint table
    if(this->gotoSubint()){
      std::cerr << 
	"Error <PSRFits::initialise> : Could not go to the SUBINT table.\n";
    }
    else this->subintTable.writeKeywords(this->fptr);
    
  }

 //--------------------------------------------------------------------------
 //--------------------------------------------------------------------------

  void Subint::writeKeywords(fitsfile *fptr)
  {
    int status=0;
    fits_update_key(fptr,TINT,"NCH_FILE",&this->nch_file,NULL,&status);
    if(status) FITSError("Subint::writeKeywords 1",status);
    
    status=0;
    fits_update_key(fptr,TSTRING,"INT_TYPE", (char *)this->int_type.c_str(),NULL,&status);
    if(status) FITSError("Subint::writeKeywords 2",status);
    
    status=0;
    fits_update_key(fptr,TSTRING,"INT_UNIT",(char *)this->int_unit.c_str(),NULL,&status);
    if(status) FITSError("Subint::writeKeywords 3",status);
    
    status=0;
    fits_update_key(fptr,TINT,"NCH_STRT",&this->nch_strt,NULL,&status);
    if(status) FITSError("Subint::writeKeywords 4",status);
    
    status=0;
    fits_update_key(fptr,TINT,"NPOL",&this->npol,NULL,&status);
    if(status) FITSError("Subint::writeKeywords 5",status);
    
    status=0;
    fits_update_key(fptr,TINT,"NBIN",&this->nbin,NULL,&status);
    if(status) FITSError("Subint::writeKeywords 6",status);
    
    status=0;
    fits_update_key(fptr,TINT,"NSBLK",&this->nsblk,NULL,&status);
    if(status) FITSError("Subint::writeKeywords 7",status);
    
    status=0;
    fits_update_key(fptr,TFLOAT,"TSAMP",&this->tsamp,NULL,&status);
    if(status) FITSError("Subint::writeKeywords 8",status);
    
    status=0;
    int colnum=0;
    fits_get_colnum(fptr, CASEINSEN, "DATA", &colnum, &status);
    if(status) FITSError("Subint::writeKeywords 10", status);

    int numData = this->nbin * this->nch_file * this->npol * this->nsblk;
    //    int dataPerByte = 8 / block.getKeyword<int>(NBITS,0);
    //    int numBytes = numData / dataPerByte;   
    int numBytes = numData / 8;
    status=0;
    if(USE_X) fits_modify_vector_len (fptr, colnum, numData, &status);
    else      fits_modify_vector_len (fptr, colnum, numBytes, &status);
    if(status) FITSError("Subint::writeKeywords 12", status);

    // delete the existing "TDIM19 = (*,*,*,*)"
    status=0;
    fits_delete_key(fptr, "TDIM19", &status);
    if(status) FITSError("Subint::writeKeywords 9", status);
    int naxis=4;
    long naxes[4];
    naxes[0] = this->nbin;
    naxes[1] = this->nch_file;
    naxes[2] = this->npol;
    naxes[3] = this->nsblk;
    std::cerr << naxes[0] << " "
	      << naxes[1] << " "
	      << naxes[2] << " "
	      << naxes[3] << " "
	      << naxes[0]*naxes[1]*naxes[2]*naxes[3] << "\n";
      
    status=0;
    fits_write_tdim(fptr, colnum, naxis, naxes, &status);
    if(status) FITSError("Subint::writeKeywords 11", status);

  }
 //--------------------------------------------------------------------------

  void PSRFits::writeSubintRow(DataBlock &block)
  {
    this->startNewSubintRow(); 
    // CHECK -- need some check of blk_cntr to make sure the number is 
    // CHECK -- incrementing correctly?
    float tsubint = block.getKeyword<float>(BLK_SEC);
    double off_sub;
    
    off_sub = 0.5*(2.*(block.getKeyword<int>(BLK_CNTR)-this->blk_initial)+1)
      * block.getKeyword<double>(BLK_SEC);

    double lst_sub  = dmsToDec(block.getKeyword<std::string>(LST));
    double ra_sub   = dmsToDec(block.getKeyword<std::string>(RA))*15.;
    double dec_sub  = dmsToDec(block.getKeyword<std::string>(DEC));
    double glon_sub = block.getKeyword<double>(GL);
    double glat_sub = block.getKeyword<double>(GB);
    float  fd_ang   = block.getKeyword<float>(FEED_ANGLE);
    float  par_ang  = block.getKeyword<float>(PANGLE);
    float  pos_ang  = remainderf(par_ang + fd_ang, 360.);
    float  tel_az   = block.getKeyword<float>(TEL_AZ);
    float  tel_zen  = block.getKeyword<float>(TEL_ZEN);
    
    this->writeColumnValue("TSUBINT",  tsubint);
    this->writeColumnValue("OFFS_SUB", off_sub);
    this->writeColumnValue("LST_SUB",  lst_sub);
    this->writeColumnValue("RA_SUB",   ra_sub);
    this->writeColumnValue("DEC_SUB",  dec_sub);
    this->writeColumnValue("GLON_SUB", glon_sub);
    this->writeColumnValue("GLAT_SUB", glat_sub);
    this->writeColumnValue("FD_ANG",   fd_ang);
    this->writeColumnValue("POS_ANG",  pos_ang);
    this->writeColumnValue("PAR_ANG",  par_ang);
    this->writeColumnValue("TEL_AZ",   tel_az);
    this->writeColumnValue("TEL_ZEN",  tel_zen);

    /////////////////////////////////////

    long nrows = this->getNumRows();
    int colnum = this->getColnum("DAT_FREQ");
    int status=0;
    fits_modify_vector_len (this->fptr, colnum, 
			    this->subintTable.nch_file, &status);
    if(status) FITSError("PSRFits::writeSubintRow", status);
    status=0;
    fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		   this->subintTable.nch_file, this->subintTable.dat_freq, 
		   &status);
    if(status) FITSError("PSRFits::writeSubintRow", status);

    colnum = this->getColnum("DAT_WTS");
    status=0;
    fits_modify_vector_len (this->fptr, colnum, 
			    this->subintTable.nch_file, &status);
    if(status) FITSError("PSRFits::writeSubintRow", status);
    status=0;
    fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		   this->subintTable.nch_file, this->subintTable.dat_wts, 
		   &status);
    if(status) FITSError("PSRFits::writeSubintRow", status);

    colnum = this->getColnum("DAT_OFFS");
    status=0;
    fits_modify_vector_len (this->fptr, colnum, 
			    this->subintTable.nch_file, &status);
    if(status) FITSError("PSRFits::writeSubintRow", status);
    status=0;
    fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		   this->subintTable.nch_file, this->subintTable.dat_offs, 
		   &status);
    if(status) FITSError("PSRFits::writeSubintRow", status);

    colnum = this->getColnum("DAT_SCL");
    status=0;
    fits_modify_vector_len (this->fptr, colnum, 
			    this->subintTable.nch_file, &status);
    if(status) FITSError("PSRFits::writeSubintRow", status);
    status=0;
    fits_write_col(this->fptr, TFLOAT, colnum, nrows, 1, 
		   this->subintTable.nch_file, this->subintTable.dat_scl, 
		   &status);
    if(status) FITSError("PSRFits::writeSubintRow", status);

    //////////////////////////////////////////

    unsigned char *data = (unsigned char *)block.getData();
    int numData = this->subintTable.nbin * this->subintTable.nch_file * 
      this->subintTable.npol * this->subintTable.nsblk;
    int dataPerByte = 8 / block.getKeyword<int>(NBITS,0);
    int numBytes = numData / dataPerByte;   
    colnum = this->getColnum("DATA");
    status=0;
    fits_write_col_byt(this->fptr, colnum, nrows, 1, numBytes, data, &status);
    if(status) FITSError("PSRFits::writeDataArray 4", status);
    

  }

}
